"""Linear model."""

import numpy as np
import pandas as pd


class _LinearModel:
    """Common abstract class for all linear models."""

    def __init__(self, fit_intercept: bool = False) -> None:
        """Initialize the model.

        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        self.fit_intercept = fit_intercept

    def fit(self, X: np.array, y: np.array) -> "_LinearModel":
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: _LinearModel
        """
        return self

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        return X @ self._coef + self._intercept

    def _assign_coef_and_intercept(self, beta: np.array):
        self._beta = beta
        if self.fit_intercept:
            self._coef = beta[1:]
            self._intercept = beta[0]
        else:
            self._coef = beta
            self._intercept = 0



"""StandardizingScaler."""

from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import norm

from remodels.transformers.BaseScaler import BaseScaler


def mad(x):
    """Compute the Mean Absolute Deviation (MAD) of an array."""
    median = np.median(x)
    dev = np.abs(x - median)
    return np.mean(dev)


class StandardizingScaler(BaseScaler):
    """A custom scaler for standardizing data using either the median or mean method.

    This scaler is suitable for preprocessing datasets in preparation for machine learning models.
    It standardizes the data, bringing it to a common scale without distorting differences in the range
    of values. The scaler can operate using either the median or mean to calculate the center and scale
    of the data.
    """

    def __init__(self, method="median") -> None:
        """Initialize the StandardizingScaler with the chosen method of centering and scaling.

        :param method: Method to use for centering ('median' or 'mean').
        :type method: str
        """
        if method not in ["median", "mean"]:
            raise ValueError("Method must be 'median' or 'mean'")
        self.method = method
        self.x_centers = None
        self.x_scales = None
        self.y_center = None
        self.y_scale = None

    def _compute_center_scale(self, data: pd.DataFrame) -> Tuple:
        """Compute the center and scale of the data based on the specified method.

        :param data: The data for which to compute the center and scale.
        :type data: pd.DataFrame
        :return: A tuple of center and scale.
        :rtype: tuple
        """
        if self.method == "median":
            center = np.median(data)
            scale = mad(data) / norm.ppf(0.75)
        else:
            center = np.mean(data)
            scale = np.std(data)

        # Avoid division by zero: if scale is 0, set it to 1
        scale = 1 if scale == 0 else scale
        return center, scale

    def fit(self, X: pd.DataFrame, y: pd.DataFrame = None) -> "StandardizingScaler":
        """Fit the scaler to the features X and optionally to the target y.

        :param X: Features to fit.
        :type X: pd.DataFrame
        :param y: Optional target to fit.
        :type y: pd.DataFrame, optional
        :return: The fitted scaler.
        :rtype: StandardizingScaler
        """
        self.x_centers, self.x_scales = self._vectorize_data(
            X, self._compute_center_scale
        )
        if y is not None:
            self.y_center, self.y_scale = self._compute_center_scale(np.array(y))
        return self

    def _vectorize_data(self, data: pd.DataFrame, func: callable) -> np.array:
        """Apply a function to each column of the data and return a vector of the results.

        :param data: The data to vectorize.
        :type data: array-like
        :param func: The function to apply to each column of the data.
        :type func: callable
        :return: A vector of the function results.
        :rtype: array-like
        """
        if isinstance(data, pd.DataFrame):
            return np.array([func(data.iloc[:, i]) for i in range(data.shape[1])]).T
        elif len(data.shape) == 1:
            return func(data)
        return np.array([func(data[:, i]) for i in range(data.shape[1])]).T

    def _apply_transform(
        self, data: pd.DataFrame, centers: pd.DataFrame, scales: pd.DataFrame
    ):
        if len(data.shape) == 1:
            return (data - centers) / scales
        return (data - centers.reshape(1, -1)) / scales.reshape(1, -1)

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform the features X and optionally the target y using the fitted scaler.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame, optional
        :return: The transformed features and optionally the transformed target.
        :rtype: tuple
        """
        X_transformed = self._apply_transform(X, self.x_centers, self.x_scales)
        if y is None:
            return X_transformed
        y_transformed = (np.array(y) - self.y_center) / self.y_scale
        return X_transformed, self._to_dataframe(y, y_transformed)

    def _apply_inverse_transform(self, data, centers, scales):
        """Apply the inverse transformation to the data.

        :param data: Transformed data to inverse transform.
        :type data: array-like
        :param centers: Centers used in the original transformation.
        :type centers: array-like
        :param scales: Scales used in the original transformation.
        :type scales: array-like
        :return: Original data before transformation.
        :rtype: array-like
        """
        if len(data.shape) == 1:
            return data * scales + centers
        return data * scales.reshape(1, -1) + centers.reshape(1, -1)

    def inverse_transform(self, X: np.array, y: np.array = None):
        """Apply the inverse transformation to the features X and optionally the target y.

        :param X: Transformed features to inverse transform.
        :type X: array-like, optional
        :param y: Transformed target to inverse transform.
        :type y: array-like, optional
        :return: The original features and target.
        :rtype: tuple
        """
        X_inverted, y_inverted = None, None
        if X is not None:
            X_inverted = self._apply_inverse_transform(X, self.x_centers, self.x_scales)
        if y is not None:
            y_inverted = (np.array(y) * self.y_scale) + self.y_center

        return (self._to_dataframe(X, X_inverted), self._to_dataframe(y, y_inverted))



"""remodels.transformers.VSTransformers."""

from .ArcsinhScaler import ArcsinhScaler
from .BoxCoxScaler import BoxCoxScaler
from .ClippingScaler import ClippingScaler
from .LogClippingScaler import LogClippingScaler
from .LogisticScaler import LogisticScaler
from .MLogScaler import MLogScaler
from .PITScaler import PITScaler
from .PolyScaler import PolyScaler



"""remodels.qra.tester."""

from .qr_results_summary import QR_ResultsSummary
from .qr_tester import QR_Tester
from .qr_tester import QR_TestResults
from .qr_tester import f_ave
from .qr_tester import q_ave



"""TimeTransformers."""

from typing import Tuple

import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class DSTAdjuster(BaseScaler):
    """A transformer for adjusting time series data to account for Daylight Saving Time (DST) changes.

    This class provides functionality to modify time series data by removing timezone information and
    resampling to an hourly frequency. It's designed to handle potential issues arising from DST transitions,
    such as duplicate or missing timestamps. The transformer can be used with any time series data that
    includes timezone-aware datetime indices.
    """

    def __init__(self):
        """Initialize the DSTAdjuster."""
        super().__init__()

    def fit(self, X: pd.DataFrame, y: pd.DataFrame = None) -> "DSTAdjuster":
        """Fit the transformer to the data.

        This transformer does not learn anything from the data
        and hence the fit method is a placeholder that returns self.

        :param X: Features to fit.
        :type X: pd.DataFrame
        :param y: Optional target to fit. Not used in this transformer.
        :type y: pd.Series, optional
        :return: The fitted transformer.
        :rtype: DSTAdjuster
        """
        # No fitting necessary for DSTAdjuster, so just return self.
        return self

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform the time series data to adjust for DST changes.

        :param X: Time series data to trasnsform.
        :type X: pd.DataFrame
        :param y: Optional target series corresponding to the time series data.
        :type y: pd.Series, optional
        :return: Adjusted time series data, and optionally the target series.
        :rtype: pd.DataFrame, pd.Series
        """
        # Remove timezone information and resample to hourly frequency.
        X_adj = X.tz_localize(None).resample("H").mean()

        # Fill missing values by averaging the adjacent values, then forward fill.
        X_adj = X_adj.fillna(X_adj.shift(-1) / 2 + X_adj.shift(1) / 2)

        if y is not None:
            # Ensure y is a pandas Series with the correct index.
            y_adj = pd.Series(y, index=X_adj.index)

            # Fill missing values in y using the same approach as for X.
            y_adj = y_adj.fillna(y_adj.shift(-1) / 2 + y_adj.shift(1) / 2)

            return X_adj, y_adj

        return X_adj



"""ArcsinhScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class ArcsinhScaler(BaseScaler):
    """A scaler that applies an arcsinh (inverse hyperbolic sine) transformation to the data.

    This scaler is useful for handling data with skewed distributions and can help in stabilizing
    the variance of the data. The transformation is stateless and does not depend on the data itself,
    meaning no fitting is required.

    The scaler also provides an inverse transformation function, which applies the sinh (hyperbolic sine)
    transformation to revert the data back to its original scale.
    """

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the arcsinh transformation to the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_transformed = np.arcsinh(X)
        y_transformed = np.arcsinh(y) if y is not None else None
        return (X_transformed, y_transformed) if y is not None else X_transformed

    def inverse_transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the inverse arcsinh transformation to the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame
        :return: Inverse transformed features and optionally inverse transformed target.
        :type: Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_inverted = np.sinh(X) if X is not None else None
        y_inverted = np.sinh(y) if y is not None else None
        return (X_inverted, y_inverted)



"""FQRA model."""

from typing import Tuple

import numpy as np

from ._linear_model import _LinearModel
from .qra import QRA


class FQRA(QRA):
    """A class that represents the FQRA model.

    In the FQRA model, factors summarizing the information in the input variables are estimated with PCA method.
    Selected number of factors is used to fit the quantile regression model.
    """

    def __init__(
        self, quantile: float = 0.5, n_factors: int = None, fit_intercept: bool = False
    ) -> None:
        """Initialize the FQRA model.

        :param quantile: quantile
        :type quantile: float
        :param n_factors: number of factors (principal components) to use; if None, number of factors is selected automatically using  Bayesian information criterion
        :type n_factors: int
        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        super().__init__(quantile, fit_intercept)
        self.n_factors = n_factors

    def fit(self, X: np.array, y: np.array):
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: FQRA
        """
        self._X_train = X
        self._y_train = y
        return self

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        X_train_f, X_test_f = self._get_factors(X)

        if self.n_factors is None:
            self.n_factors = self._select_best_n_factors_with_bic(
                X_train_f,
                QRA(self.quantile, self.fit_intercept),
            )

        X_train_f = X_train_f[:, : self.n_factors]
        X_test_f = X_test_f[:, : self.n_factors]
        qra = QRA(self.quantile, self.fit_intercept).fit(X_train_f, self._y_train)
        return qra.predict(X_test_f)

    def _get_factors(self, X: np.array) -> Tuple[np.array, np.array]:
        X_full = np.concatenate([self._X_train, X], axis=0)
        try:
            _, _, Vh = np.linalg.svd(X_full, full_matrices=False)
            F = (X_full @ Vh.T) / X.shape[1]
        except np.linalg.LinAlgError:
            F = X_full
        return F[: self._X_train.shape[0], :], F[self._X_train.shape[0] :, :]

    def _select_best_n_factors_with_bic(
        self,
        X_train_f: np.array,
        model: _LinearModel,
    ) -> int:
        last_bic = np.inf
        best_n_factors = 0
        N, K = X_train_f.shape

        for i in range(1, K + 1):
            model = model.fit(X_train_f[:, :i], self._y_train)
            y_hat_train = model.predict(X_train_f[:, :i])
            rss = np.sum(np.square(self._y_train - y_hat_train))
            bic = N * np.log(rss / N) + i * np.log(N)
            if bic > last_bic:
                break
            last_bic = bic
            best_n_factors = i
        return best_n_factors



"""sFQRA model."""

from typing import Tuple

import numpy as np

from .fqra import FQRA


class sFQRA(FQRA):
    """A class that represents the FQRA model.

    The sFQRA model is an FQRA model where the input variables are standardized by subtracting the mean and dividing by the standard deviation (calculated across rows).
    """

    def __init__(
        self, quantile: float = None, n_factors: int = None, fit_intercept: bool = False
    ) -> None:
        """Initialize the sFQRA model.

        :param quantile: quantile
        :type quantile: float
        :param n_factors: number of factors (principal components) to use; if None, number of factors is selected automatically using  Bayesian information criterion
        :type n_factors: int
        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        super().__init__(quantile, n_factors, fit_intercept)

    def fit(self, X: np.array, y: np.array):
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: sFQRA
        """
        X, mean, std = self._zscore(X)
        y = (y - mean) / std
        return super().fit(X, y)

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        X, mean, std = self._zscore(X)
        y = super().predict(X)
        return y * std + mean

    def _zscore(self, X: np.array) -> Tuple[np.array, np.array, np.array]:
        mean = np.mean(X, axis=1)
        std = np.std(X, axis=1)
        return (X - mean[:, np.newaxis]) / std[:, np.newaxis], mean, std



"""FQRM model."""

import numpy as np

from ._lstsq import _LSTSQ
from .fqra import FQRA
from .qra import QRA


class FQRM(FQRA):
    """A class that represents the FQRM model.

    In the FQRM model, factors summarizing the information in the input variables are estimated with PCA method.
    The average of the calculated factors is used to fit the quantile regression model.
    """

    def __init__(
        self, quantile: float = 0.5, n_factors: int = None, fit_intercept: bool = False
    ) -> None:
        """Initialize FQRM model.

        :param quantile: quantile
        :type quantile: float
        :param n_factors: number of factors (principal components) to use; if None, number of factors is selected automatically using  Bayesian information criterion
        :type n_factors: int
        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        super().__init__(quantile, n_factors, fit_intercept)

    def fit(self, X: np.array, y: np.array):
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: FQRM
        """
        return super().fit(X, y)

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        X_train_f, X_test_f = self._get_factors(X)

        if self.n_factors is None:
            self.n_factors = self._select_best_n_factors_with_bic(
                X_train_f,
                _LSTSQ(self.fit_intercept),
            )

        X_train_f = X_train_f[:, : self.n_factors]
        X_test_f = X_test_f[:, : self.n_factors]

        lstsq = _LSTSQ(self.fit_intercept).fit(X_train_f, self._y_train)
        X_train_fm = lstsq.predict(X_train_f)
        X_train_fm = np.expand_dims(X_train_fm, axis=1)
        X_test_fm = lstsq.predict(X_test_f)
        X_test_fm = np.expand_dims(X_test_fm, axis=1)

        qra = QRA(self.quantile, self.fit_intercept).fit(X_train_fm, self._y_train)
        return qra.predict(X_test_fm)



"""QRM model."""

import numpy as np

from .qra import QRA


class QRM(QRA):
    """A class that represents the QRM model.

    In the QRM model, the average of the input variables is first calculated.
    This average is used to fit the quantile regression model.
    """

    def __init__(self, quantile: float = 0.5, fit_intercept: bool = False) -> None:
        """Initialize the QRM model.

        :param quantile: quantile
        :type quantile: float
        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        super().__init__(quantile=quantile, fit_intercept=fit_intercept)

    def fit(self, X: np.array, y: np.array):
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: QRM
        """
        X = np.mean(X, axis=1, keepdims=True)
        return super().fit(X, y)

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        X = np.mean(X, axis=1, keepdims=True)
        return super().predict(X)



"""BaseScaler."""

from typing import Tuple

import pandas as pd


class BaseScaler:
    """Custom scaler base class following scikit-learn's conventions."""

    def fit(self, X, y=None) -> "BaseScaler":
        """Fit the scaler to the data. Placeholder that does nothing.

        :param X: Input data.
        :type X: array-like
        :param y: Optional, target values (None by default).
        :type y: array-like, optional
        :return: Returns self.
        :rtype: BaseScaler
        """
        # No fitting process required for base scaler.
        return self

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transforms the data. Placeholder that should be overridden by subclasses.

        :param X: Input data to transform.
        :type X: array-like
        :param y: Optional, target values (None by default).
        :type y: array-like, optional
        :return: Transformed data.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        # BaseScaler does not implement transform method.
        raise NotImplementedError("Transform method not implemented.")

    def _to_dataframe(
        self, original: pd.DataFrame, transformed: pd.DataFrame
    ) -> pd.DataFrame:
        """Converts transformed data back to a DataFrame if the original was a DataFrame.

        :param original: Original input data.
        :type original: array-like
        :param transformed: Transformed data.
        :type transformed: array-like
        :return: Transformed data as a DataFrame if original was a DataFrame, otherwise array-like.
        :rtype: pd.DataFrame
        """
        if isinstance(original, pd.DataFrame):
            return pd.DataFrame(
                transformed, index=original.index, columns=original.columns
            )
        else:
            return transformed

    def fit_transform(self, X: pd.DataFrame, y: pd.DataFrame = None) -> pd.DataFrame:
        """Fit to data, then transform it.

        :param X: Features to fit and transform.
        :type X: np.ndarray
        :param y: Optional target to fit and transform.
        :type y: np.ndarray, optional
        :return: Transformed features and optionally transformed target.
        :rtype: tuple or np.ndarray
        """
        # Call the fit method (even if it does nothing in this case)
        self.fit(X, y)

        # Call the transform method and return its result
        return self.transform(X, y)

    def inverse_transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Inverse transformat the data. Placeholder that should be overridden by subclasses.

        :param X: Input data to transform.
        :type X: array-like
        :param y: Optional, target values (None by default).
        :type y: array-like, optional
        :return: Transformed data.
        :rtype: array-like
        """
        # BaseScaler does not implement transform method.
        raise NotImplementedError("Transform method not implemented.")



"""PointModel."""

import datetime as dt

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score

from remodels.pipelines.RePipeline import RePipeline


class PointModel:
    """PointModel is a time-series prediction model designed to forecast electricity prices or similar data.

    This model is equipped with a flexible data processing pipeline and the ability to handle different sets
    of predictor variables for different hours of the day. It offers functionality for model training,
    prediction with a rolling window approach, and calculation of various evaluation metrics.
    """

    def __init__(
        self, pipeline: RePipeline, variables_per_hour: dict = {}, y_column="price_da"
    ):
        """Initialize the PointModel with a data processing pipeline, variables mapped to each hour, and the target column name.

        :param pipeline: Sequence of data transformation steps and a predictive model.
        :type pipeline: RePipeline
        :param variables_per_hour: Mapping from hour ranges to the variables to be used in those hours.
        :type variables_per_hour: dict
        :param y_column: The name of the target column.
        :type y_column: str

        """
        self.variables_per_hour = variables_per_hour
        self.transformation_pipeline = pipeline[:-1]
        self.all_used_columns = list(
            np.unique(
                [
                    item
                    for sublist in self.variables_per_hour.values()
                    for item in sublist
                ]
            )
        )
        self.model = pipeline[-1]
        self.unique_hours = None
        self.y_column = [y_column]
        self.predictions = None
        self.training_data = None

    def fit(self, df: pd.DataFrame, start: str, end: str):
        """Fit the model with the training data.

        :param df: DataFrame containing the training data.
        :type df: pd.DataFrame
        :param start: start of fitting
        :type start: str
        :param end: end of fitting
        :type end: str
        """
        self.training_data = df
        self.set_unique_hours(df.index)
        self.start = start
        self.end = end

    def set_unique_hours(self, dates: pd.Series):
        """Set the unique hours for the model based on the provided datetime data.

        :param dates: Datetime data to extract unique hours from.
        :type dates: pd.Series
        """
        if self.unique_hours is None:
            self.unique_hours = np.sort(dates.hour.unique())

    def get_hour_variables(self, hour: int):
        """Retrieve the variables associated with a specific hour based on defined hour ranges.

        :param hour: The hour for which variables are needed.
        :type hour: int
        :return: List of variables associated with the given hour.
        :rtype: list
        """
        for hour_range, variables in self.variables_per_hour.items():
            if hour_range[0] <= hour <= hour_range[1]:
                return variables

    def separate_columns_by_dtype(self, df: pd.DataFrame):
        """Separate columns in a DataFrame by data type (float vs non-float).

        :param df: DataFrame to separate columns from.
        :type df: pandas.DataFrame
        :return: Lists of float columns and non-float columns.
        :rtype: tuple
        """
        float_columns = df.select_dtypes(include=[np.float64]).columns.tolist()
        non_float_columns = [col for col in df.columns if col not in float_columns]
        return float_columns, non_float_columns

    def fit_transform_data(self, Xy: pd.DataFrame, is_train: bool = True):
        """Fit the transformation pipeline to the data and transform it if is_train is True, otherwise, only transform the data.

        :param Xy: DataFrame containing features and target to be transformed.
        :type Xy: pd.DataFrame
        :param is_train: Flag to indicate whether to fit the transformer or not.
        :type is_train: bool
        :return: Transformed features and optionally transformed target.
        :rtype: tuple or pd.DataFrame
        """
        float_columns, non_float_columns = self.separate_columns_by_dtype(
            Xy[self.all_used_columns]
        )
        non_float_data = Xy[non_float_columns]

        if is_train:
            X, y = self.transformation_pipeline.fit_transform(
                Xy[float_columns], Xy[self.y_column]
            )
            X = pd.concat([X, non_float_data], axis=1)
            return X, y
        else:
            X_transformed = self.transformation_pipeline.transform(Xy[float_columns])
            X = pd.concat([X_transformed, non_float_data], axis=1)
            return X

    def train_and_predict_hours(
        self,
        day: pd.Timestamp,
        Xy_train: pd.DataFrame,
        Xy_test: pd.DataFrame,
        predictions_list: list,
        inverse_predictions: bool,
    ):
        """Train the model and make predictions for each hour in the unique_hours, and store the predictions in a list.

        :param day: The day for which predictions are made.
        :type day: pd.Timestamp
        :param Xy_train: Training data.
        :type Xy_train: pd.DataFrame
        :param Xy_test: Testing data.
        :type Xy_test: pd.DataFrame
        :param predictions_list: List to store the predictions.
        :type predictions_list: list
        :param inverse_predictions: Flag to determine whether to apply inverse transformation to predictions.
        :type inverse_predictions: bool
        """
        for hour in self.unique_hours:
            hour_variables = self.get_hour_variables(hour)
            if hour_variables is None:
                hour_variables = self.all_used_columns
            Xy_train_hour = Xy_train.loc[Xy_train.index.hour == hour]
            X_test_hour = Xy_test.loc[Xy_test.index.hour == hour][hour_variables]
            Xy_train_hour, y_train_hour = self.fit_transform_data(
                Xy_train_hour[hour_variables + self.y_column]
            )
            X_test_hour = self.fit_transform_data(
                X_test_hour[hour_variables], is_train=False
            )
            date_ = pd.Timestamp(f"{day} {hour}:00:00")

            if len(Xy_train_hour) > 0 and len(X_test_hour) > 0:
                self.model.fit(Xy_train_hour, y_train_hour)
                prediction = self.model.predict(X_test_hour)

                if (len(prediction) > 0) and inverse_predictions:
                    _, prediction_df = self.transformation_pipeline.inverse_transform(
                        yt=prediction[0]
                    )
                    predictions_list.append((date_, prediction_df[0]))
                elif len(prediction) > 0:
                    predictions_list.append((date_, prediction[0, 0]))

    def predict(self, calibration_window: int = 728, inverse_predictions: bool = True):
        """Predict values over a given range, from start to end, using a rolling window, and store/update predictions in the model.

        :param df: DataFrame containing the data to be used for prediction.
        :type df: pd.DataFrame
        :param calibration_window: Number of days to look back for training data.
        :type calibration_window: int
        :param inverse_predictions: Flag to determine whether to apply inverse transformation to predictions.
        :type inverse_predictions: bool
        :return: DataFrame of predicted values.
        :rtype: pd.DataFrame
        """
        df = self.training_data
        if not len(self.all_used_columns):
            self.all_used_columns = df.drop(columns=self.y_column).columns.tolist()
        predictions_list = []
        for day in pd.date_range(self.start, self.end, freq="D"):
            Xy_train = df.loc[
                (df.index.date >= day.date() - dt.timedelta(days=calibration_window))
                & (df.index.date < day.date())
            ][self.all_used_columns + self.y_column].dropna()
            Xy_test = df.loc[df.index.date == day.date()][
                self.all_used_columns + self.y_column
            ].dropna()
            self.train_and_predict_hours(
                day, Xy_train, Xy_test, predictions_list, inverse_predictions
            )

        new_predictions_df = pd.DataFrame(
            predictions_list, columns=["DateTime", f"prediction_{calibration_window}rw"]
        )
        new_predictions_df.set_index("DateTime", inplace=True)

        if (
            self.predictions is not None
            and new_predictions_df.columns[0] not in self.predictions.columns
        ):
            self.predictions = self.predictions.join(new_predictions_df, how="outer")
        else:
            self.predictions = new_predictions_df

        return self.predictions[[f"prediction_{calibration_window}rw"]]

    def calculate_metrics(self, y_true: pd.DataFrame, y_pred: pd.DataFrame):
        """Calculate regression metrics.

        :param y_true: DataFrame containing the actual data
        :type y_pred: pd.DataFrame
        :param y_pred: DataFrame containing the predicted data
        :type y_pred: pd.DataFrame
        :return: dict of calculated regression metrics
        :rtype: dict
        """
        mae = mean_absolute_error(y_true, y_pred)
        mse = mean_squared_error(y_true, y_pred)
        rmse = np.sqrt(mse)
        mape = np.mean(np.abs((y_true - y_pred) / y_true)) * 100
        r2 = r2_score(y_true, y_pred)
        return {"MAE": mae, "MSE": mse, "RMSE": rmse, "MAPE": mape, "R2": r2}

    def summary(self):
        """Generate a summary comparing stored predictions with actual values from the training data.

        :return: DataFrame with summary metrics.
        :rtype: pd.DataFrame
        """
        if self.predictions is None:
            raise ValueError("No predictions have been made yet.")
        if self.training_data is None:
            raise ValueError("Model has not been fitted with training data yet.")

        aligned_df = self.training_data[self.y_column].join(
            self.predictions, how="inner"
        )
        actual_values = aligned_df[self.y_column[0]]
        metrics = {}
        for col in self.predictions.columns:
            if col.startswith("prediction_"):
                metrics[col] = self.calculate_metrics(actual_values, aligned_df[col])

        metrics_df = pd.DataFrame(metrics).T

        return metrics_df



"""BoxCoxScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class BoxCoxScaler(BaseScaler):
    """A scaler that applies a Box-Cox transformation to the data.

    The Box-Cox transformation is a statistical technique used to stabilize variance, make the data
    more normally distributed, and improve the validity of measures of association. It's particularly
    effective for transforming non-normal dependent variables into a normal shape. The transformation
    is defined as:

    Y(λ) = (X^λ - 1) / λ, if λ != 0
           log(X), if λ = 0

    where X is the original data and λ is the transformation parameter. The λ value is chosen to
    maximize the normality of the transformed data. A λ of 0 implies a log transformation, while
    other values indicate various degrees of exponential transformation.

    This scaler includes both the Box-Cox transformation and its inverse, enabling reversible scaling
    of data.
    """

    def __init__(self, lamb: float = 0.5) -> None:
        """Initialize the scaler with a lambda parameter for the Box-Cox transformation.

        :param lamb: Lambda parameter for the Box-Cox transformation.
        :type lamb: float
        """
        self.lamb = lamb

    def _transform_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """Apply the Box-Cox transformation to the data.

        :param data: Data to transform.
        :type data: pd.DataFrame or np.ndarray
        :return: Transformed data.
        :rtype: pd.DataFrame
        """
        if self.lamb != 0:
            transformed = np.sign(data) * (
                ((np.abs(data) + 1) ** self.lamb - 1) / self.lamb
            )
        else:
            transformed = np.log(np.abs(data) + 1)
        return transformed

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the transformation to the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_transformed = self._transform_data(X)
        y_transformed = self._transform_data(y) if y is not None else None
        return (X_transformed, y_transformed) if y is not None else X_transformed

    def inverse_transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the inverse Box-Cox transformation to the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame
        :return: Inverse transformed features and optionally inverse transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """

        def invert(data):
            return (
                np.sign(data) * ((self.lamb * np.abs(data) + 1) ** (1 / self.lamb) - 1)
                if self.lamb != 0
                else np.exp(np.abs(data)) - 1
            )

        X_inverted = invert(X) if X is not None else None
        y_inverted = invert(y) if y is not None else None
        return (X_inverted, y_inverted)



"""*QR* tester."""

import concurrent.futures
import pickle
from itertools import repeat
from typing import Callable
from typing import Iterable

import numpy as np
from scipy.stats import chi2
from tqdm.auto import tqdm

from remodels.qra import QRA


def _process(X_train, y_train, X_test, q, qra_model):
    qra_model.quantile = q / 100
    y_test = qra_model.fit(X_train, y_train).predict(X_test)
    return q, y_test


_default_qr_model = QRA(fit_intercept=True)


class QR_Tester:
    """QR Tester is a class for testing QR models.

    QR Tester class is a class designed to obtain probabilistic predictions using a given QR model.

    The QR model is fitted to the data portion specified by the calibration window. Then, the QR model predictions for all percentiles are calculated. The process is repeated for subsequent portions of data.
    """

    def __init__(
        self,
        calibration_window: int = 7 * 24,
        prediction_window: int = 24,
        multivariate: bool = True,
        qr_model=_default_qr_model,
        max_workers: int = None,
        progress: bool = True,
    ) -> None:
        """Initialize the QR Tester.

        :param calibration_window: length of calibration window, defaults to 7 * 24
        :type calibration_window: int, optional
        :param prediction_window: length of prediction window, defaults to 24
        :type prediction_window: int, optional
        :param qr_model: *QR* model, defaults to QRA(fit_intercept=True)
        :type qr_model: QRA, optional
        :param max_workers: process pool executor max workers, defaults to None
        :type max_workers: int, optional
        """
        self.calibration_window = calibration_window
        self.prediction_window = prediction_window
        self.multivatiate = multivariate
        self.qr_model = qr_model
        self.max_workers = max_workers
        self.progress = progress

    def fit_predict(self, X: np.array, y: np.array) -> "QR_TestResults":
        """Run QR Tester to obtain probabilistic predictions wrapped in special results class.

        :param X: data matrix
        :type X: np.array
        :param y: endogenous variable
        :type y: np.array
        :return: QR_TestResults object
        :rtype: QR_TestResults
        """
        executor = concurrent.futures.ProcessPoolExecutor(self.max_workers)

        Y_pred = np.zeros((X.shape[0] - self.calibration_window, 99), np.float_)

        for i in tqdm(
            range(
                0,
                X.shape[0] - self.calibration_window,
                self.prediction_window,
            ),
            disable=not self.progress,
        ):
            X_train = X[i : i + self.calibration_window]
            y_train = y[i : i + self.calibration_window]
            X_test = X[
                i
                + self.calibration_window : i
                + self.calibration_window
                + self.prediction_window
            ]
            if self.multivatiate:
                for h in range(self.prediction_window):
                    X_train_h = X_train[h :: self.prediction_window]
                    y_train_h = y_train[h :: self.prediction_window]
                    X_test_h = X_test[h :: self.prediction_window]
                    for q, y_test in executor.map(
                        _process,
                        repeat(X_train_h),
                        repeat(y_train_h),
                        repeat(X_test_h),
                        range(1, 100),
                        repeat(self.qr_model),
                    ):
                        Y_pred[i + h, q - 1] = y_test
            else:
                for q, y_test in executor.map(
                    _process,
                    repeat(X_train),
                    repeat(y_train),
                    repeat(X_test),
                    range(1, 100),
                    repeat(self.qr_model),
                ):
                    Y_pred[i : i + self.prediction_window, q - 1] = y_test

        Y_pred = np.sort(Y_pred, axis=1)

        return QR_TestResults(
            Y_pred,
            y[self.calibration_window :],
            self.prediction_window,
        )


def q_ave(*results: Iterable["QR_TestResults"]):
    """Results quantile averaging."""
    for r in results:
        assert r.Y_pred.shape == results[0].Y_pred.shape
        assert np.array_equal(r.y_test, results[0].y_test)
        assert r.prediction_window == results[0].prediction_window

    Y_pred_qave = np.array([r.Y_pred for r in results]).mean(axis=0)

    return QR_TestResults(
        Y_pred=Y_pred_qave,
        y_test=results[0].y_test,
        prediction_window=results[0].prediction_window,
    )


def f_ave(*results: Iterable["QR_TestResults"]):
    """Results probability averaging."""
    for r in results:
        assert r.Y_pred.shape == results[0].Y_pred.shape
        assert np.array_equal(r.y_test, results[0].y_test)
        assert r.prediction_window == results[0].prediction_window

    Y_pred_fave = np.zeros_like(results[0].Y_pred)

    for k in range(Y_pred_fave.shape[0]):
        rows = [r.Y_pred[k, :] for r in results]
        linspace = np.sort(np.concatenate(rows))

        q = np.zeros_like(linspace)

        for idx, x in enumerate(linspace):
            q[idx] = np.mean([(row <= x).sum() for row in rows])

        q = (np.ceil(q) - 1).astype(int)

        fave = np.zeros(shape=(99,))
        for i in range(99):
            slice_ = linspace[q == i]
            if slice_.shape[0] != 0:
                fave[i] = np.mean(slice_)
        for i in range(99):
            if fave[i] == 0:
                if i == 0:
                    fave[i] = fave[i + 1]
                elif i == 98:
                    fave[i] = fave[i - 1]
                else:
                    fave[i] = (fave[i - 1] + fave[i + 1]) / 2

        Y_pred_fave[k, :] = fave

    return QR_TestResults(
        Y_pred=Y_pred_fave,
        y_test=results[0].y_test,
        prediction_window=results[0].prediction_window,
    )


class QR_TestResults:
    """A class that wraps probabilistic predictions.

    The QR Test Results allows you to calculate metric values.
    """

    def __init__(
        self,
        Y_pred: np.array,
        y_test: np.array,
        prediction_window: int,
    ) -> None:
        """Initialize the QR Test Results class.

        :param Y_pred: matrix of predictions
        :type Y_pred: np.array
        :param y_test: endogenous variable
        :type y_test: np.array
        :param prediction_window: length of prediction window
        :type prediction_window: int
        """
        self.Y_pred = Y_pred
        self.y_test = y_test
        self.prediction_window = prediction_window

    def to_pickle(self, file):
        """Save QR Test Results to pickle.

        :param file: pickle file
        :type file: FileDescriptorOrPath
        """
        with open(file, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def read_pickle(file):
        """Read QR Test Results from pickle.

        :param file: pickle file
        :type file: FileDescriptorOrPath
        :return: QR Test Results object
        :rtype: QR_TestResults
        """
        with open(file, "rb") as f:
            qr_test_results = pickle.load(f)
        return qr_test_results

    def aec(self, alpha: int) -> float:
        """Average empirical coverage.

        :param alpha: length of prediction interval
        :type alpha: int
        :return: average empirical coverage value
        :rtype: float
        """
        hits = self._hits(alpha)
        return hits.sum() / self.Y_pred.shape[0]

    def ec_h(self, alpha: int) -> np.array:
        """Empirical coverage per 'hour'.

        :param alpha: length of prediction interval
        :type alpha: int
        :return: emipirical coverage per hour values
        :rtype: np.array
        """
        hits = self._hits(alpha)
        ec_h = np.zeros(shape=(self.prediction_window,))
        for i in range(self.prediction_window):
            hits_h = hits[i :: self.prediction_window]
            ec_h[i] = hits_h.sum() / hits_h.shape[0]
        return ec_h

    def ec_mad(self, alpha: int) -> float:
        """Empirical coverage per 'hour' - mean absolute deviation.

        :param alpha: length of prediction interval
        :type alpha: int
        :return: mean absolute deviation of empirical coverage per hour values
        :rtype: float
        """
        return np.mean(np.abs(self.ec_h(alpha) - alpha / 100))

    def kupiec_test(self, alpha: int, significance_level: float = 0.05) -> int:
        """Kupiec test. Count the number of times the null hypothesis is not rejected.

        :param alpha: length of predition interval
        :type alpha: int
        :param significance_level: test significance level, defaults to 0.05
        :type significance_level: float, optional
        :return: number of hours that test is not rejected
        :rtype: int
        """
        return self._count_hypothesis_not_rejected(
            self._kupiec_test_statistic, alpha, significance_level, 1
        )

    def christoffersen_test(self, alpha: int, significance_level: float = 0.05) -> int:
        """Christoffersen test. Count the number of times the null hypothesis is not rejected.

        :param alpha: length of predition interval
        :type alpha: int
        :param significance_level: test significance level, defaults to 0.05
        :type significance_level: float, optional
        :return: number of hours that test is not rejected
        :rtype: int
        """
        return self._count_hypothesis_not_rejected(
            self._christoffersen_test_statistic, alpha, significance_level, 2
        )

    def _count_hypothesis_not_rejected(
        self,
        test_func: Callable[[np.array, float], float],
        alpha: int,
        significance_level: float,
        df: int,
    ) -> int:
        alpha_p = alpha / 100
        hits = self._hits(alpha)
        hypothesis_not_rejected_counter = 0

        for i in range(self.prediction_window):
            hits_h = hits[i :: self.prediction_window]
            test_statistic = test_func(hits_h, alpha_p)
            is_not_rejected = test_statistic < chi2.ppf(1 - significance_level, df=df)
            hypothesis_not_rejected_counter += is_not_rejected

        return hypothesis_not_rejected_counter

    def _kupiec_test_statistic(self, hits: np.array, alpha_p: float) -> float:
        n = hits.shape[0]
        n1 = hits.sum()
        n0 = n - n1
        L_0 = n1 * np.log(alpha_p) + n0 * np.log(1 - alpha_p)
        L_A = n1 * np.log(n1 / n) + n0 * np.log(n0 / n)
        return 2 * (L_A - L_0)

    def _christoffersen_test_statistic(self, hits: np.array, alpha_p: float) -> float:
        t00 = (~hits & ~self.__shift_arr(hits, -1)).sum()
        t01 = (~hits & self.__shift_arr(hits, -1)).sum()
        t10 = (hits & ~self.__shift_arr(hits, -1)).sum()
        t11 = (hits & self.__shift_arr(hits, -1)).sum()

        p01 = t01 / (t00 + t01)
        p11 = t11 / (t11 + t10)
        L_A = (
            t00 * np.log(1 - p01)
            + t01 * np.log(p01)
            + t10 * np.log(1 - p11)
            + t11 * np.log(p11)
        )
        L_0 = (t00 + t10) * np.log(1 - alpha_p) + (t01 + t11) * np.log(alpha_p)
        return 2 * (L_A - L_0)

    @staticmethod
    def __shift_arr(arr, num, fill_value=np.nan):
        result = np.empty_like(arr)
        if num > 0:
            result[:num] = fill_value
            result[num:] = arr[:-num]
        elif num < 0:
            result[num:] = fill_value
            result[:num] = arr[-num:]
        else:
            result[:] = arr
        return result

    def _hits(self, alpha) -> np.array:
        lower_bound = 49 - alpha // 2
        upper_bound = 49 + alpha // 2
        return (self.Y_pred[:, lower_bound] <= self.y_test) & (
            self.y_test <= self.Y_pred[:, upper_bound]
        )

    def aps(self) -> float:
        """Aggregate pinball score.

        :return: aggregate pinball score value
        :rtype: float
        """
        return np.mean(self._pinball_score_matrix())

    def aps_extreme_quantiles(self, n_quantiles: int) -> float:
        """Aggregate pinball score for n extreme quantiles.

        :param n_quantiles: number of leftmost and rightmost quantiles
        :type n_quantiles: int
        :return: aggregate pinball score computed for extreme quantiles
        :rtype: float
        """
        pinball_score_matrix = self._pinball_score_matrix()
        idx = list(range(n_quantiles)) + list(range(99 - n_quantiles, 99))
        return np.mean(pinball_score_matrix[:, idx])

    def _pinball_score_matrix(self) -> np.array:
        q = np.arange(0.01, 1.00, 0.01)
        q = np.expand_dims(q, axis=0)
        q = np.repeat(q, self.Y_pred.shape[0], axis=0)

        resid = self.y_test[:, np.newaxis] - self.Y_pred
        return (q - (resid < 0)) * resid



"""PITScaler."""

from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.stats import t

from remodels.transformers.BaseScaler import BaseScaler


class PITScaler(BaseScaler):
    """Probability Integral Transform (PIT) Scaler applies a transformation to data based on a specified probability distribution.

    This scaler transforms each feature using the cumulative distribution function (CDF) of the specified distribution,
    effectively mapping the empirical CDF of the data to the target distribution. This technique is often used
    in statistical modeling and forecasting to normalize data or make it conform to a certain distribution.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def __init__(self, distribution: str = "normal", nu: int = 8) -> None:
        """Initialize the PIT-Scaler.

        :param distribution: distribution, defaults to "normal"
        :type distribution: str, optional
        :param nu: distribution parameter, defaults to 8
        :type nu: int, optional

        """
        self.distribution = distribution
        self.nu = nu
        self.empirical_cdfs = {}
        self.y_column_name = None

    def fit(self, X: pd.DataFrame, y: pd.DataFrame = None) -> "PITScaler":
        """Fit the scaler to the data.

        :param X: Input data.
        :type X: pd.DataFrame
        :param y: Optional, target values (None by default).
        :type y: pd.DataFrame, optional
        :return: Returns self.
        :rtype: PITScaler

        """
        if y is not None:
            y_name = y.columns[0]
            self.y_column_name = y_name
            y = pd.DataFrame(y, columns=[y_name])
            X = pd.concat([X, y], axis=1)

        for column in X.columns:
            sorted_data = np.sort(X[column])
            self.empirical_cdfs[column] = (
                sorted_data,
                np.arange(1, len(sorted_data) + 1) / len(sorted_data),
            )
        return self

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transforms the data.

        :param X: Input data to transform.
        :type X: pd.DataFrame
        :param y: Optional, target values (None by default).
        :type y: pd.DataFrame, optional
        :return: Transformed data.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]

        """
        transformed_data = pd.DataFrame(index=X.index)
        for column in X.columns:
            sorted_data, empirical_cdf = self.empirical_cdfs[column]
            pit_values = np.interp(X[column], sorted_data, empirical_cdf)
            epsilon = 1e-10  # A small tolerance value
            pit_values = np.clip(pit_values, epsilon, 1 - epsilon)
            transformed_data[column] = self._apply_distribution(pit_values)

        if y is not None and self.y_column_name in self.empirical_cdfs:
            sorted_data, empirical_cdf = self.empirical_cdfs[self.y_column_name]
            pit_values = np.interp(y, sorted_data, empirical_cdf)
            epsilon = 1e-10  # A small tolerance value
            pit_values = np.clip(pit_values, epsilon, 1 - epsilon)
            y_transformed = self._apply_distribution(pit_values)
            y_transformed = self._to_dataframe(y, y_transformed)
            return transformed_data, y_transformed

        return transformed_data

    def _apply_distribution(self, pit_values):
        if self.distribution == "normal":
            return norm.ppf(pit_values)
        elif self.distribution == "student-t":
            return t.ppf(pit_values, df=self.nu)
        else:
            raise ValueError("Invalid distribution type. Use 'normal' or 'student-t'.")

    def inverse_transform(
        self, X: pd.DataFrame = None, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Inverse transform the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame, optional
        :return: Original features and optionally original target.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]

        """
        if X is not None:
            inverted_data = pd.DataFrame(index=X.index)
            for column in X.columns:
                sorted_data, empirical_cdf = self.empirical_cdfs[column]
                transformed_values = self._apply_inverse_distribution(X[column])
                inverted_data[column] = np.interp(
                    transformed_values, empirical_cdf, sorted_data
                )
        else:
            inverted_data = None

        if y is not None and self.y_column_name in self.empirical_cdfs:
            sorted_data, empirical_cdf = self.empirical_cdfs[self.y_column_name]
            transformed_values = self._apply_inverse_distribution(y)
            y_inverted = np.interp(transformed_values, empirical_cdf, sorted_data)
            y_inverted = self._to_dataframe(y, y_inverted)
        else:
            y_inverted = None
        return inverted_data, y_inverted

    def _apply_inverse_distribution(self, transformed_values):
        if self.distribution == "normal":
            return norm.cdf(transformed_values)
        elif self.distribution == "student-t":
            return t.cdf(transformed_values, df=self.nu)
        else:
            raise ValueError("Invalid distribution type. Use 'normal' or 'student-t'.")



"""LogisticScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class LogisticScaler(BaseScaler):
    """Scaler that applies a logistic transformation to the data. This transformation converts each feature using the logistic function, which maps any real-valued number into the range (0, 1).

    The transformation is particularly useful in preparing data for algorithms that expect input values
    to be in a bounded range. It can also help in dealing with features that have skewed distributions.

    The logistic transformation is defined as:
        1 / (1 + exp(-x))

    where 'x' is the feature value.

    The scaler provides both the transformation and its inverse, allowing the original scale of the data
    to be recovered.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the logistic transformation to the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame, optional
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_transformed = (1 + np.exp(-X)) ** (-1)
        y_transformed = (1 + np.exp(-y)) ** (-1) if y is not None else None
        return (X_transformed, y_transformed) if y is not None else X_transformed

    def inverse_transform(
        self, X: pd.DataFrame = None, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the inverse logistic transformation to the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame, optional
        :return: Original features and optionally original target.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]
        """

        def invert(data):
            return np.log(data / (1 - data))

        X_inverted = invert(X) if X is not None else None
        y_inverted = invert(y) if y is not None else None
        return (X_inverted, y_inverted)



"""LogClippingScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class LogClippingScaler(BaseScaler):
    """Scaler that applies a logarithmic transformation to values exceeding a specified threshold.

    This scaler is designed to transform features by applying a logarithmic transformation, but only to
    values that exceed a certain threshold, 'k'. This approach can be particularly useful in reducing the
    impact of outliers or extreme values in the data, while maintaining the scale of the rest of the data.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def __init__(self, k=3):
        """Initialize the scaler with a clipping threshold.

        :param k: Clipping threshold.
        :type k: float
        """
        self.k = k

    def _transform_data(self, data):
        """Apply the log clipping transformation to the data.

        :param data: Data to transform.
        :type data: pd.DataFrame or np.ndarray
        :return: Transformed data.
        :rtype: pd.DataFrame or np.ndarray
        """
        condition = np.abs(data) > self.k
        data_transformed = data.copy()
        data_transformed = np.maximum(data_transformed.abs(), 2 + 1e-8)
        data_transformed = np.where(
            condition, np.sign(data) * (np.log(np.abs(data_transformed) - 2) + 3), data
        )
        return data_transformed

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the transformation to the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame, optional
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_transformed = self._transform_data(X)
        y_transformed = self._transform_data(y) if y is not None else None
        return (
            (self._to_dataframe(X, X_transformed), self._to_dataframe(y, y_transformed))
            if y is not None
            else self._to_dataframe(X, X_transformed)
        )

    def inverse_transform(
        self, X: pd.DataFrame = None, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Apply the inverse log clipping transformation to the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame, optional
        :return: Original features and optionally original target.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]
        """

        def invert(data):
            condition = np.abs(data) > self.k
            data_inverted = np.where(
                condition, np.sign(data) * (np.exp(np.abs(data) - 3) + 2), data
            )
            return data_inverted

        X_inverted = invert(X) if X is not None else None
        y_inverted = invert(y) if y is not None else None
        return (self._to_dataframe(X, X_inverted), self._to_dataframe(y, y_inverted))



"""ClippingScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class ClippingScaler(BaseScaler):
    """Scaler that clips feature and target values to within a specified number of standard deviations from the mean.

    This scaler limits extreme values in the data by clipping them to a defined range based on a
    multiple of standard deviations. It is particularly useful for mitigating the effect of outliers
    in the data, making it more robust for various statistical analyses or machine learning models.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def __init__(self, k=3) -> None:
        """Initialize the ClippingScaler with a clipping threshold.

        :param k: The number of standard deviations to use as the clipping threshold.
        :type k: int or float
        """
        self.k = k

    def fit(self, X: pd.DataFrame, y: pd.DataFrame = None) -> "ClippingScaler":
        """Fit the scaler to the data.

        This scaler does not learn anything from the data
        and hence the fit method is a placeholder that returns self.

        :param X: Features to fit.
        :type X: np.ndarray or pd.DataFrame
        :param y: Optional target to fit. Not used in this scaler.
        :type y: np.ndarray or pd.Series, optional
        :return: The fitted scaler.
        :rtype: ClippingScaler
        """
        # No fitting necessary for ClippingScaler, so just return self.
        return self

    def _clip_data(self, data):
        """Clip the data to within the threshold defined by self.k.

        :param data: The data to clip.
        :type data: np.ndarray or pd.Series
        :return: The clipped data.
        :rtype: np.ndarray
        """
        if isinstance(data, pd.Series):
            data = data.values
        condition = np.abs(data) > self.k
        return np.where(condition, self.k * np.sign(data), data)

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform the features and optionally the target by clipping their values.

        :param X: Features to transform.
        :type X: np.ndarray or pd.DataFrame
        :param y: Optional target to transform.
        :type y: np.ndarray or pd.Series, optional
        :return: The transformed features and optionally the transformed target.
        :rtype: tuple
        """
        X_transformed = self._clip_data(X)
        y_transformed = self._clip_data(y) if y is not None else None
        return (
            (self._to_dataframe(X, X_transformed), self._to_dataframe(y, y_transformed))
            if y is not None
            else self._to_dataframe(X, X_transformed)
        )

    def inverse_transform(
        self, X: pd.DataFrame = None, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Inverse transform the features and optionally the target by unclipping their values.

        This method assumes the original data was within the range [-k, k].

        :param X: Transformed features to inverse transform.
        :type X: np.ndarray or pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: np.ndarray or pd.Series, optional
        :return: The original features and target.
        :rtype: tuple
        """
        X_inverted = np.clip(X, -self.k, self.k) if X is not None else None
        y_inverted = np.clip(y, -self.k, self.k) if y is not None else None
        return (self._to_dataframe(X, X_inverted), self._to_dataframe(y, y_inverted))



"""QR Results Summary."""

from typing import Callable
from typing import Dict
from typing import List

import pandas as pd

from .qr_tester import QR_TestResults


class QR_ResultsSummary:
    """A class to summarize QR Test Results."""

    table_styles = [
        {"selector": ".red", "props": "background-color: #fcb8b8;"},
        {"selector": ".green", "props": "background-color: #b8fcc3;"},
        {"selector": ".bold", "props": "font-weight: bold;"},
    ]

    def __init__(self, results_dict: Dict[str, Dict[str, QR_TestResults]]) -> None:
        """Initialize the QR Results Summary class.

        :param results_dict: dictionary of results by dataset and QR variant
        :type results_dict: Dict[str, Dict[str, QR_TestResults]]
        """
        self.results_dict = results_dict
        self.dataset_names = results_dict.keys()

        self.qr_variants = []
        for dataset_results in results_dict.values():
            for qr_variant_name in dataset_results.keys():
                if qr_variant_name not in self.qr_variants:
                    self.qr_variants.append(qr_variant_name)

    def aec(self, alpha_list: List[int]) -> pd.DataFrame:
        """Average empirical coverage.

        :param alpha_list: length of prediction interval list
        :type alpha_list: List[int]
        :return: aec summary
        :rtype: pd.DataFrame
        """
        df = self._alpha_dataset_variant_summary(alpha_list, QR_TestResults.aec)
        style_df = self._get_style_df(df)

        for alpha in alpha_list:
            for dataset_name in self.dataset_names:
                for qr_variant in self.qr_variants:
                    value = df.at[qr_variant, (alpha, dataset_name)] * 100
                    if abs(alpha - value) < 1:
                        style_df.at[qr_variant, (alpha, dataset_name)] += "green "
                    elif abs(alpha - value) > 2.5:
                        style_df.at[qr_variant, (alpha, dataset_name)] += "red "
                    if (
                        abs(value - alpha)
                        == (df.loc[:, (alpha, dataset_name)] * 100 - alpha).abs().min()
                    ):
                        style_df.at[qr_variant, (alpha, dataset_name)] += "bold "

        return (
            df.astype(float)
            .applymap(lambda x: f"{(x*100):.2f}")
            .style.set_table_styles(self.table_styles, overwrite=True)
            .set_td_classes(style_df)
            .set_caption("Average empirical coverage")
        )

    def kupiec_test(self, alpha_list: List[int]) -> pd.DataFrame:
        """Kupiec test.

        :param alpha_list: length of prediction interval list
        :type alpha_list: List[int]
        :return: Kupiec test summary
        :rtype: pd.DataFrame
        """
        df = self._alpha_dataset_variant_summary(alpha_list, QR_TestResults.kupiec_test)
        style_df = self._get_style_df(df)

        for alpha in alpha_list:
            for dataset_name in self.dataset_names:
                for qr_variant in self.qr_variants:
                    value = df.at[qr_variant, (alpha, dataset_name)]
                    if value >= 20:
                        style_df.at[qr_variant, (alpha, dataset_name)] += "green "
                    elif value < 12:
                        style_df.at[qr_variant, (alpha, dataset_name)] += "red "

        return (
            df.astype(int)
            .style.set_table_styles(self.table_styles, overwrite=True)
            .set_td_classes(style_df)
            .set_caption("Kupiec test")
        )

    def _alpha_dataset_variant_summary(self, alpha_list: List[int], func: Callable):
        df = pd.DataFrame(
            index=self.qr_variants,
            columns=pd.MultiIndex.from_product(
                [alpha_list, self.dataset_names], names=["alpha", "dataset"]
            ),
        )

        for alpha in alpha_list:
            for dataset_name in self.dataset_names:
                for qr_variant in self.qr_variants:
                    df.at[qr_variant, (alpha, dataset_name)] = func(
                        self.results_dict[dataset_name][qr_variant], alpha
                    )

        return df

    def aps(self) -> pd.DataFrame:
        """Aggregate pinball score.

        :return: aggregate pinball score summary
        :rtype: pd.DataFrame
        """
        return self._dataset_variant_summary(
            func=QR_TestResults.aps, caption="Aggregate pinball score"
        )

    def aps_extreme_quantiles(self, n_quantiles: int) -> pd.DataFrame:
        """Aggregate pinball score for n extreme quantiles.

        :param n_quantiles: number of leftmost and rightmost quantiles
        :type n_quantiles: int
        :return: aggregate pinball score computed for extreme quantiles summary
        :rtype: pd.DataFrame
        """
        return self._dataset_variant_summary(
            func=QR_TestResults.aps_extreme_quantiles,
            caption=f"Aggregate pinball score<br>{2*n_quantiles} extreme quantiles",
            n_quantiles=n_quantiles,
        )

    def _dataset_variant_summary(self, func: Callable, caption: str, **kwargs):
        df = pd.DataFrame(index=self.qr_variants, columns=self.dataset_names)

        for dataset_name in self.dataset_names:
            for qr_variant in self.qr_variants:
                df.at[qr_variant, dataset_name] = func(
                    self.results_dict[dataset_name][qr_variant], **kwargs
                )

        style_df = self._get_style_df(df)

        for dataset_name in self.dataset_names:
            for qr_variant in self.qr_variants:
                value = df.at[qr_variant, dataset_name]
                if value == df.loc[:, dataset_name].min():
                    style_df.at[qr_variant, dataset_name] += "green "

        return (
            df.astype(float)
            .style.format(precision=3)
            .set_table_styles(
                [
                    {"selector": ".green", "props": "background-color: #b8fcc3;"},
                ],
                overwrite=True,
            )
            .set_td_classes(style_df)
            .set_caption(caption)
        )

    def _get_style_df(self, df: pd.DataFrame) -> pd.DataFrame:
        style_df = df.copy()
        style_df.loc[:, :] = ""
        return style_df



"""QRA-computing helper functions."""

import numpy as np
from scipy import sparse
from scipy.optimize import linprog
from scipy.optimize import minimize
from scipy.stats import iqr
from scipy.stats import norm


def _add_intercept(X: np.array) -> np.array:
    new_X = np.ones(shape=(X.shape[0], X.shape[1] + 1))
    new_X[:, 1:] = X
    return new_X


def _lqra(
    X, y, quantile: float = 0.5, lambda_: float = 0.0, fit_intercept: bool = False
):
    # input matrix X - N rows x K columns
    N, K = X.shape

    if fit_intercept:
        K += 1
        X = _add_intercept(X)

    # scaling regularization parameter to make it sample-independent
    lambda_ = N * lambda_

    c = np.concatenate(
        [
            np.full((2 * K,), fill_value=lambda_),
            np.full((N,), fill_value=quantile),
            np.full((N,), fill_value=1 - quantile),
        ]
    )
    if fit_intercept:  # do not penalize intercept
        c[0] = 0
        c[K] = 0

    b_eq = y
    # A_eq = np.concatenate([X, -X, np.eye(N), -np.eye(N)], axis=1)

    eye = sparse.eye(N, dtype=X.dtype, format="csc")
    A_eq = sparse.hstack([X, -X, eye, -eye], format="csc")

    optimize_result = linprog(
        c=c,
        A_eq=A_eq,
        b_eq=b_eq,
        method="highs",  # highs is an optimalized solver (simplex or interior-point)
    )

    # beta = beta+ - beta-
    beta = optimize_result.x[:K] - optimize_result.x[K : 2 * K]
    return beta


def _sqra(X, y, quantile: float = 0.5, H: float = None, fit_intercept=False):
    # SQRA can be solved with gradient-descent method (2022-he)
    # for now, TNC (non-linear) minimizer is used

    if fit_intercept:
        X = _add_intercept(X)

    beta__initial_guess = _lqra(X, y, quantile=quantile, lambda_=0, fit_intercept=False)
    if H is None:  # rule-of-thumb estimation
        residuals = y - (X @ beta__initial_guess)
        resid_std = np.std(residuals)
        resid_iqr = iqr(residuals)
        H = min(resid_std, resid_iqr) * 1.06 * (X.shape[0] ** (-1 / 5))
    H = max(H, 1e-4)

    def rho(beta):
        residuals = y - X @ beta
        return (
            np.sum(H * norm.pdf(residuals / H))
            + (quantile - norm.cdf(-residuals / H)) @ residuals
        )

    def rho_der(beta):
        residuals = y - X @ beta
        return (X.T @ (norm.cdf(-residuals / H) - quantile)) / X.shape[0]

    def rho_hess(beta):
        return X.T @ (X * (norm.pdf((X @ beta - y) / H) / H)[np.newaxis].T) / X.shape[0]

    return minimize(
        rho,
        beta__initial_guess,
        method="Newton-CG",
        jac=rho_der,
        hess=rho_hess,
        options=dict(xtol=1e-5),
    ).x



"""remodels.qra."""

from .fqra import FQRA
from .fqrm import FQRM
from .lqra import LQRA
from .qra import QRA
from .qrm import QRM
from .sfqra import sFQRA
from .sfqrm import sFQRM
from .sqra import SQRA
from .sqrm import SQRM
from .tester import QR_ResultsSummary
from .tester import QR_Tester
from .tester import QR_TestResults
from .tester import f_ave
from .tester import q_ave



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['remodels',
 'remodels.data',
 'remodels.pipelines',
 'remodels.pointsModels',
 'remodels.qra',
 'remodels.qra.tester',
 'remodels.transformers',
 'remodels.transformers.VSTransformers']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.0,<3.8.0',
 'pandas<=1.5.3',
 'scikit-learn>=1.3.1,<2.0.0',
 'tqdm>=4.66.1,<5.0.0']

entry_points = \
{'console_scripts': ['remodels = remodels.__main__:main']}

setup_kwargs = {
    'name': 'remodels',
    'version': '1.0.1',
    'description': 'remodels',
    'long_description': "# remodels\n\n[![PyPI](https://img.shields.io/pypi/v/remodels.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/remodels.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/remodels)][python version]\n[![License](https://img.shields.io/pypi/l/remodels)][license]\n\n[![Documentation Status](https://readthedocs.org/projects/remodels/badge/?version=latest)][read the docs]\n[![Tests](https://github.com/zakrzewow/remodels/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/zakrzewow/remodels/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/remodels/\n[status]: https://pypi.org/project/remodels/\n[python version]: https://pypi.org/project/remodels\n[read the docs]: https://remodels.readthedocs.io/en/latest/\n[tests]: https://github.com/zakrzewow/remodels/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/zakrzewow/remodels\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\nReModels is a Python package for probabilistic energy price forecasting using Quantile Regression Averaging (QRA) methods.\n\n## Installation\n\nYou can install _remodels_ via [pip] from [PyPI]:\n\n```console\n$ pip install remodels\n```\n\nAlternatively, you can install from source:\n\n```console\n$ git clone https://github.com/zakrzewow/remodels.git\n$ cd remodels\n$ pip install .\n```\n\n## Features\n\n- **Dataset Download**: access commonly used public datasets for transparent data acquisition.\n- **Data Preprocessing**: apply variance stabilizing transformation for improved data quality.\n- **Forecast Generation**: produce point and probabilistic forecasts with reference implementations of QRA variants.\n- **Result Evaluation**: compare predictions using dedicated metrics for fair and consistent evaluation.\n\nReModels provides a robust framework for researchers to compare different QRA methods and other forecasting techniques. It supports the development of new forecasting methods, extending beyond energy price forecasting.\n\nReModels simplifies and enhances energy price forecasting research with comprehensive tools and transparent methodologies.\n\nThe main functionality of the ReModels package are Variance Stabilizing Transformations and Quantile Regression Averaging variants.\n\nImplemented VSTs:\n\n- clipping,\n- clipping with the logarithm function\n- logistic transform\n- inverse hyperbolic sine transform\n- BoxCox transform\n- polynomial transform\n- mirror-log transformation\n- probability integral transform\n\nImplemented QRA variants:\n\n- QRA\n- QRM\n- FQRA\n- FQRM\n- sFQRA\n- sFQRM\n- LQRA\n- SQRA\n- SQRM\n\n## Usage\n\nVST basic usage example:\n\n```py\nimport numpy as np\nfrom remodels.transformers.VSTransformers import ArcsinhScaler, BoxCoxScaler, ClippingScaler, LogClippingScaler, LogisticScaler, MLogScaler, PITScaler, PolyScaler\n\nX = np.random.normal(0, 1, size=(100, 10))\nArcsinhScaler().fit_transform(X)\n```\n\nQRA basic usage example:\n\n```py\nimport numpy as np\nfrom remodels.qra import QRA, QRM, FQRA, FQRM, sFQRA, sFQRM, LQRA, SQRA, SQRM\n\nX = np.random.normal(0, 1, size=(100, 10))\ny = X.sum(axis=1)\n\nqra = QRA(quantile=0.5, fit_intercept=True)\nqra.fit(X, y).predict(X)\n```\n\nFor more advanced examples of high-level functionalities, please see the [Usage] or [Reference].\n\n## License\n\nDistributed under the terms of the [MIT license][license],\n_remodels_ is free and open source software.\n\n## Credits\n\nThis project was generated from [@cjolowicz]'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/project/remodels/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[pip]: https://pip.pypa.io/\n\n<!-- github-only -->\n\n[license]: https://github.com/zakrzewow/remodels/blob/main/LICENSE\n[usage]: https://remodels.readthedocs.io/en/latest/usage.html\n[reference]: https://remodels.readthedocs.io/en/latest/reference.html\n",
    'author': 'Grzegorz Zakrzewski',
    'author_email': 'zakrzew12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zakrzewow/remodels',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""sFQRM model."""

import numpy as np

from .fqrm import FQRM
from .sfqra import sFQRA


class sFQRM(FQRM, sFQRA):
    """A class that represents the FQRA model.

    The sFQRM model is an FQRM model where the input variables are standardized by subtracting the mean and dividing by the standard deviation (calculated across rows).
    """

    def __init__(
        self, quantile: float = None, n_factors: int = None, fit_intercept: bool = False
    ) -> None:
        """Initialize the sFQRM model.

        :param quantile: quantile
        :type quantile: float
        :param n_factors: number of factors (principal components) to use; if None, number of factors is selected automatically using  Bayesian information criterion
        :type n_factors: int
        :param fit_intercept: True if fit intercept in model, defaults to False
        :type fit_intercept: bool, optional
        """
        super().__init__(quantile, n_factors, fit_intercept)

    def fit(self, X: np.array, y: np.array):
        """Fit the model to the data.

        :param X: input matrix
        :type X: np.array
        :param y: dependent variable
        :type y: np.array
        :return: fitted model
        :rtype: sFQRM
        """
        return sFQRA.fit(self, X, y)

    def predict(self, X: np.array) -> np.array:
        """Predict the dependent variable.

        :param X: input matrix
        :type X: np.array
        :return: prediction
        :rtype: np.array
        """
        X, mean, std = self._zscore(X)
        y = FQRM.predict(self, X)
        return y * std + mean



"""PolyScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class PolyScaler(BaseScaler):
    r"""Scaler that applies a polynomial transformation to data, transforming each feature according to a polynomial function.

    The transformation is defined as:

        sign(x) * ((\|x\| + (c / lamb)^(1 / (lamb - 1)))^lamb - (c / lamb)^(lamb / (lamb - 1)))

    where 'lamb' is the exponent parameter, and 'c' is a constant determining the curvature of
    the polynomial. This transformation can be particularly useful for stabilizing variance and
    making skewed distributions more symmetric.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def __init__(self, lamb: float = 0.125, c: float = 0.05) -> None:
        """Initialize the scaler with parameters for the polynomial transformation.

        :param lamb: Exponent used in the polynomial transformation.
        :type lamb: float
        :param c: Constant that defines the curvature of the polynomial transformation.
        :type c: float

        """
        self.lamb = lamb
        self.c = c

    def _transform_data(self, data: np.ndarray) -> np.ndarray:
        """Apply the polynomial transformation to the data.

        :param data: Data to transform.
        :type data: np.ndarray
        :return: Transformed data.
        :rtype: np.ndarray

        """
        return np.sign(data) * (
            (data.abs() + (self.c / self.lamb) ** (1 / (self.lamb - 1))).abs()
            ** self.lamb
            - (self.c / self.lamb) ** (self.lamb / (self.lamb - 1))
        )

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame, optional
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]

        """
        X_transformed = self._transform_data(X)
        return (
            (X_transformed, self._transform_data(y)) if y is not None else X_transformed
        )

    def inverse_transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Inverse transform the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: np.ndarray
        :param y: Transformed target to inverse transform.
        :type y: np.ndarray, optional
        :return: Original features and optionally original target.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]

        """

        def invert(data):
            c_lamb = (self.c / self.lamb) ** (self.lamb / (self.lamb - 1))
            return np.sign(data) * (
                (np.abs(data) + c_lamb) ** (1 / self.lamb) - c_lamb ** (1 / (self.lamb))
            )

        X_inverted = invert(X) if X is not None else None
        y_inverted = invert(y) if y is not None else None
        return (X_inverted, y_inverted)



"""MLogScaler."""

from typing import Tuple

import numpy as np
import pandas as pd

from remodels.transformers.BaseScaler import BaseScaler


class MLogScaler(BaseScaler):
    r"""Scaler that applies a modified logarithmic transformation to the data. This transformation is designed to handle zero and negative values effectively by incorporating a small constant.

    The transformation is defined as:
        sign(x) * (log(\|x\| + 1/c) + log(c))

    where 'c' is a small constant to ensure non-zero division. This transformation helps in
    stabilizing variance and normalizing distributions, especially useful for skewed data.

    The scaler also provides an inverse transformation function to revert the data back to
    its original scale.
    """

    def __init__(self, c: int = 1 / 3):
        """Initialize the scaler with a constant used in the transformation.

        :param c: A small constant to ensure non-zero division in transformation.
        :type c: float
        """
        self.c = c

    def _transform_data(self, data: pd.DataFrame):
        """Apply the modified logarithmic transformation to the data.

        :param data: Data to transform.
        :type data: pd.DataFrame
        :return: Transformed data.
        :rtype: pd.DataFrame
        """
        return np.sign(data) * (np.log(np.abs(data) + 1 / self.c) + np.log(self.c))

    def transform(
        self, X: pd.DataFrame, y: pd.DataFrame = None
    ) -> pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]:
        """Transform the features and optionally the target.

        :param X: Features to transform.
        :type X: pd.DataFrame
        :param y: Optional target to transform.
        :type y: pd.DataFrame, optional
        :return: Transformed features and optionally transformed target.
        :rtype: pd.DataFrame or Tuple[pd.DataFrame, pd.DataFrame]
        """
        X_transformed = self._transform_data(X)
        return (
            (X_transformed, self._transform_data(y)) if y is not None else X_transformed
        )

    def inverse_transform(
        self, X: pd.DataFrame = None, y: pd.DataFrame = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Inverse transform the features and optionally the target.

        :param X: Transformed features to inverse transform.
        :type X: pd.DataFrame
        :param y: Transformed target to inverse transform.
        :type y: pd.DataFrame, optional
        :return: Original features and optionally original target.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]
        """

        def invert(data):
            return np.sign(data) * (np.exp(np.abs(data) - np.log(self.c)) - 1 / self.c)

        X_inverted = invert(X) if X is not None else None
        y_inverted = invert(y) if y is not None else None
        return (self._to_dataframe(X, X_inverted), self._to_dataframe(y, y_inverted))
